@@ -43,6 +43,13 @@
 #include "jhash.h"
 #include "error.h"
 
+/* Test compiler support for weak symbols with hidden visibility. */
+int __tracepoint_test_symbol1 __attribute__((weak, visibility("hidden")));
+void *__tracepoint_test_symbol2 __attribute__((weak, visibility("hidden")));
+struct {
+	char a[24];
+} __tracepoint_test_symbol3 __attribute__((weak, visibility("hidden")));
+
 /* Set to 1 to enable tracepoint debug output */
 static const int tracepoint_debug;
 static int initialized;
@@ -803,11 +810,34 @@ int tracepoint_unregister_lib(struct lttng_ust_tracepoint * const *tracepoints_s
 	return 0;
 }
 
+/*
+ * Report in debug message whether the compiler correctly supports weak
+ * hidden symbols. This test checks that the address associated with two
+ * weak symbols with hidden visibility is the same when declared within
+ * two compile units part of the same module.
+ */
+static void check_weak_hidden(void)
+{
+	DBG("Your compiler treats weak symbols with hidden visibility for integer objects as %s between compile units part of the same module.",
+		&__tracepoint_test_symbol1 == lttng_ust_tp_check_weak_hidden1() ?
+			"SAME address" :
+			"DIFFERENT addresses");
+	DBG("Your compiler treats weak symbols with hidden visibility for pointer objects as %s between compile units part of the same module.",
+		&__tracepoint_test_symbol2 == lttng_ust_tp_check_weak_hidden2() ?
+			"SAME address" :
+			"DIFFERENT addresses");
+	DBG("Your compiler treats weak symbols with hidden visibility for 24-byte structure objects as %s between compile units part of the same module.",
+		&__tracepoint_test_symbol3 == lttng_ust_tp_check_weak_hidden3() ?
+			"SAME address" :
+			"DIFFERENT addresses");
+}
+
 void init_tracepoint(void)
 {
 	if (uatomic_xchg(&initialized, 1) == 1)
 		return;
 	init_usterr();
+	check_weak_hidden();
 }
 
 void exit_tracepoint(void)
