@@ -446,6 +446,12 @@ void lib_ring_buffer_check_deliver(const struct lttng_ust_lib_ring_buffer_config
 									handle),
 					      handle);
 
+			/*
+			 * Increment the packet counter while we have exclusive
+			 * access.
+			 */
+			subbuffer_inc_packet_count(config, &buf->backend, idx, handle);
+
 			/*
 			 * Set noref flag and offset for this subbuffer id.
 			 * Contains a memory barrier that ensures counter stores
